@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: UTF-8 -*-
 #
-# Copyright 2016-2021 European Commission (JRC);
+# Copyright 2016-2022 European Commission (JRC);
 # Licensed under the EUPL (the 'Licence');
 # You may not use this work except in compliance with the Licence.
 # You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
@@ -249,6 +249,7 @@ class ExcelModel:
         if stack is None:
             stack = set(self.dsp.data_nodes) - done - set(self.references)
         stack = sorted(stack)
+        sheet_limits = {}
         while stack:
             n_id = stack.pop()
             if isinstance(n_id, sh.Token) or n_id in done:
@@ -268,7 +269,7 @@ class ExcelModel:
 
             try:
                 context = self.add_book(book)[1]
-                worksheet, context = self.add_sheet(rng['sheet'], context)
+                wk, context = self.add_sheet(rng['sheet'], context)
             except Exception as ex:  # Missing excel file or sheet.
                 log.warning('Error in loading `{}`:\n{}'.format(n_id, ex))
                 Cell(n_id, '=#REF!').compile().add(self.dsp)
@@ -279,9 +280,22 @@ class ExcelModel:
             formula_references = self.formula_references(context)
             formula_ranges = self.formula_ranges(context)
             external_links = self.external_links(context)
-            rng = '{c1}{r1}:{c2}{r2}'.format(**rng)
-            for c in flatten(worksheet[rng], None):
-                if hasattr(c, 'value'):
+
+            _name = '%s'
+            if 'sheet_id' in rng:
+                _name = f'{rng["sheet_id"]}!{_name}'
+            if wk not in sheet_limits:
+                sheet_limits[wk] = wk.max_row, wk.max_column
+            max_row, max_column = sheet_limits[wk]
+            it = wk.iter_rows(
+                int(rng['r1']), min(int(rng['r2']), max_row),
+                rng['n1'], min(rng['n2'], max_column)
+            )
+            for c in flatten(it, None):
+                n = _name % c.coordinate
+                if n in self.cells:
+                    continue
+                elif hasattr(c, 'value'):
                     cell = self.add_cell(
                         c, context, references=references,
                         formula_references=formula_references,
@@ -292,40 +306,70 @@ class ExcelModel:
                         stack.extend(cell.inputs or ())
         return self
 
-    def _assemble_ranges(self, cells, nodes=None):
-        get = sh.get_nested_dicts
-        pred = self.dsp.dmap.pred
+    def _assemble_ranges(self, cells, nodes=None, compact=1):
+        get, dsp = sh.get_nested_dicts, self.dsp
+        pred = dsp.dmap.pred
         if nodes is None:
-            nodes = set(self.dsp.data_nodes).difference(self.dsp.default_values)
+            nodes = set(dsp.data_nodes).difference(dsp.default_values)
         it = (
             k for k in nodes
             if not pred[k] and not isinstance(k, sh.Token)
         )
+        ranges = []
         for n_id in it:
             if isinstance(n_id, sh.Token):
                 continue
             try:
-                ra = RangesAssembler(n_id)
+                ra = RangesAssembler(n_id, compact=compact)
             except ValueError:
                 continue
             rng = ra.range.ranges[0]
-            for output, indices in get(cells, rng['sheet_id'], default=list):
-                if not ra.push(output, indices):
+            for out, idx in get(cells, 'range', rng['sheet_id'], default=list):
+                if not ra.push(idx, out):
                     break
-            ra.add(self.dsp)
+            ra.push(get(cells, 'cell', rng['sheet_id']))
+            ranges.append(ra)
 
-    def assemble(self):
+        for ra in sorted(ranges, key=lambda x: len(x.missing)):
+            ra.add(dsp)
+
+    def assemble(self, compact=1):
         cells, get = {}, sh.get_nested_dicts
         for c in self.cells.values():
             if isinstance(c, Ref):
                 continue
             rng = c.range.ranges[0]
-            get(cells, rng['sheet_id'], default=list).append((
-                c.output, RangesAssembler._range_indices(c.range)
-            ))
-        self._assemble_ranges(cells)
+            indices = RangesAssembler._range_indices(c.range)
+            if len(indices) == 1:
+                get(cells, 'cell', rng['sheet_id'])[list(indices)[0]] = c.output
+            else:
+                get(cells, 'range', rng['sheet_id'], default=list).append(
+                    (c.output, indices)
+                )
+
+        self._assemble_ranges(cells, compact=compact)
         return self
 
+    def inverse_references(self):
+        dsp = self.dsp
+        pred, succ, nodes = dsp.dmap.pred, dsp.dmap.succ, dsp.nodes
+        for c in tuple(self.cells.values()):
+            if isinstance(c, Ref) and c.inputs:
+                if c.func.dsp.function_nodes:
+                    continue
+                inp = c.output
+                if set(pred[inp]) == {c.func.function_id}:
+                    out = list(c.inputs)[0]
+                    if not any(out in succ[k] for k in succ[inp]):
+                        dsp.add_function(
+                            '=%s' % inp, sh.bypass, inputs=[inp], outputs=[out]
+                        )
+                        d = nodes[inp]
+                        d['inv-ref'] = out
+                        sh.get_nested_dicts(d, 'filters', default=list).extend(
+                            nodes[out].get('filters', ())
+                        )
+
     def finish(self, complete=True, circular=False, assemble=True):
         if complete:
             self.complete()
@@ -333,6 +377,7 @@ class ExcelModel:
             self.assemble()
         if circular:
             self.solve_circular()
+        self.inverse_references()
         return self
 
     def to_dict(self):
@@ -351,11 +396,11 @@ class ExcelModel:
                 nodes.update(dict.fromkeys(d['outputs'], fun.__name__))
         return nodes
 
-    def from_dict(self, adict, context=None, assemble=True):
+    def from_dict(self, adict, context=None, assemble=True, ref=True):
         refs, cells, nodes, get = {}, {}, set(), sh.get_nested_dicts
         for k, v in adict.items():
             try:
-                cell = Cell(k, v, context=context)
+                cell = Cell(k, v, context=context, replace_missing_ref=ref)
             except ValueError:
                 ref = Ref(k, v, context=context).compile(context=context)
                 nodes.update(ref.add(self.dsp))
@@ -433,9 +478,14 @@ class ExcelModel:
 
     def compile(self, inputs, outputs):
         dsp = self.dsp.shrink_dsp(outputs=outputs)
-
+        inp = set(inputs)
+        inp.update({dsp.nodes.get(i, {}).get('inv-ref', None) for i in inputs})
+        pred, start = dsp.dmap.pred, (sh.START,)
+        dsp.dmap.remove_nodes_from({
+            i for k in inp for i in pred.get(k, ()) if tuple(pred[i]) == start
+        })
         dsp.default_values = sh.selector(
-            set(dsp.default_values) - set(inputs), dsp.default_values
+            set(dsp.default_values) - inp, dsp.default_values
         )
 
         res = dsp()
